@@ -11,6 +11,7 @@ import re
 
 if python_version_tuple()[0] < "3":
     from itertools import izip_longest
+    from functools import partial
     _none_type = type(None)
     _int_type = int
     _float_type = float
@@ -18,7 +19,7 @@ if python_version_tuple()[0] < "3":
     _binary_type = str
 else:
     from itertools import zip_longest as izip_longest
-    from functools import reduce
+    from functools import reduce, partial
     _none_type = type(None)
     _int_type = int
     _float_type = float
@@ -26,8 +27,8 @@ else:
     _binary_type = bytes
 
 
-__all__ = ["tabulate"]
-__version__ = "0.6"
+__all__ = ["tabulate", "tabulate_formats", "simple_separated_format"]
+__version__ = "0.7"
 
 
 Line = namedtuple("Line", ["begin", "hline", "sep", "end"])
@@ -36,37 +37,98 @@ Line = namedtuple("Line", ["begin", "hline", "sep", "end"])
 DataRow = namedtuple("DataRow", ["begin", "sep", "end"])
 
 
+# A table structure is suppposed to be:
+#
+#     --- lineabove ---------
+#         headerrow
+#     --- linebelowheader ---
+#         datarow
+#     --- linebewteenrows ---
+#     ... (more datarows) ...
+#     --- linebewteenrows ---
+#         last datarow
+#     --- linebelow ---------
+#
+# TableFormat's line* elements can be
+#
+#   - either None, if the element is not used,
+#   - or a Line tuple,
+#   - or a function: [col_widths], [col_alignments] -> string.
+#
+# TableFormat's *row elements can be
+#
+#   - either None, if the element is not used,
+#   - or a DataRow tuple,
+#   - or a function: [cell_values], [col_widths], [col_alignments] -> string.
+#
+# padding (an integer) is the amount of white space around data values.
+#
+# with_header_hide:
+#
+#   - either None, to display all table elements unconditionally,
+#   - or a list of elements not to be displayed if the table has column headers.
+#
 TableFormat = namedtuple("TableFormat", ["lineabove", "linebelowheader",
                                          "linebetweenrows", "linebelow",
                                          "headerrow", "datarow",
-                                         "padding", "usecolons", "usehtmlattrs",
-                                         "with_header_hide",
-                                         "without_header_hide"])
+                                         "padding", "with_header_hide"])
 
 
-_format_defaults = {"padding": 0,
-                    "usecolons": False,
-                    "usehtmlattrs": False,
-                    "with_header_hide": [],
-                    "without_header_hide": []}
+def _pipe_segment_with_colons(align, colwidth):
+    """Return a segment of a horizontal line with optional colons which
+    indicate column's alignment (as in `pipe` output format)."""
+    w = colwidth
+    if align in ["right", "decimal"]:
+        return ('-' * (w - 1)) + ":"
+    elif align == "center":
+        return ":" + ('-' * (w - 2)) + ":"
+    elif align == "left":
+        return ":" + ('-' * (w - 1))
+    else:
+        return '-' * w
+
+
+def _pipe_line_with_colons(colwidths, colaligns):
+    """Return a horizontal line with optional colons to indicate column's
+    alignment (as in `pipe` output format)."""
+    segments = [_pipe_segment_with_colons(a, w) for a, w in zip(colaligns, colwidths)]
+    return u"|" + u"|".join(segments) + u"|"
+
+
+def _mediawiki_row_with_attrs(separator, cell_values, colwidths, colaligns):
+    alignment = { "left":    '',
+                  "right":   'align="right"| ',
+                  "center":  'align="center"| ',
+                  "decimal": 'align="right"| ' }
+    # hard-coded padding _around_ align attribute and value together
+    # rather than padding parameter which affects only the value
+    values_with_attrs = [' ' + alignment[a] + c + ' '
+                         for c, a in zip(cell_values, colaligns)]
+    colsep = separator*2
+    return (separator + colsep.join(values_with_attrs)).rstrip()
+
+
+def _latex_line_begin_tabular(colwidths, colaligns):
+    alignment = { "left": "l", "right": "r", "center": "c", "decimal": "r" }
+    tabular_columns_fmt = "".join([alignment[a] for a in colaligns])
+    return "\\begin{tabular}{" + tabular_columns_fmt + "}\n\hline"
 
 
 _table_formats = {"simple":
-                  TableFormat(lineabove=None,
+                  TableFormat(lineabove=Line("", "-", "  ", ""),
                               linebelowheader=Line("", "-", "  ", ""),
                               linebetweenrows=None,
                               linebelow=Line("", "-", "  ", ""),
                               headerrow=DataRow("", "  ", ""),
                               datarow=DataRow("", "  ", ""),
                               padding=0,
-                              usecolons=False,
-                              usehtmlattrs=False,
-                              with_header_hide=["linebelow"],
-                              without_header_hide=[]),
+                              with_header_hide=["lineabove", "linebelow"]),
                   "plain":
-                  TableFormat(None, None, None, None,
-                              DataRow("", "  ", ""), DataRow("", "  ", ""),
-                              **_format_defaults),
+                  TableFormat(lineabove=None, linebelowheader=None,
+                              linebetweenrows=None, linebelow=None,
+                              headerrow=DataRow("", "  ", ""),
+                              datarow=DataRow("", "  ", ""),
+                              padding=0, with_header_hide=None),
                   "grid":
                   TableFormat(lineabove=Line("+", "-", "+", "+"),
                               linebelowheader=Line("+", "=", "+", "+"),
@@ -74,23 +136,16 @@ _table_formats = {"simple":
                               linebelow=Line("+", "-", "+", "+"),
                               headerrow=DataRow("|", "|", "|"),
                               datarow=DataRow("|", "|", "|"),
-                              padding=1,
-                              usecolons=False,
-                              usehtmlattrs=False,
-                              with_header_hide=[],
-                              without_header_hide=["linebelowheader"]),
+                              padding=1, with_header_hide=None),
                   "pipe":
-                  TableFormat(lineabove=None,
-                              linebelowheader=Line("|", "-", "|", "|"),
+                  TableFormat(lineabove=_pipe_line_with_colons,
+                              linebelowheader=_pipe_line_with_colons,
                               linebetweenrows=None,
                               linebelow=None,
                               headerrow=DataRow("|", "|", "|"),
                               datarow=DataRow("|", "|", "|"),
                               padding=1,
-                              usecolons=True,
-                              usehtmlattrs=False,
-                              with_header_hide=[],
-                              without_header_hide=[]),
+                              with_header_hide=["lineabove"]),
                   "orgtbl":
                   TableFormat(lineabove=None,
                               linebelowheader=Line("|", "-", "+", "|"),
@@ -98,11 +153,7 @@ _table_formats = {"simple":
                               linebelow=None,
                               headerrow=DataRow("|", "|", "|"),
                               datarow=DataRow("|", "|", "|"),
-                              padding=1,
-                              usecolons=False,
-                              usehtmlattrs=False,
-                              with_header_hide=[],
-                              without_header_hide=["linebelowheader"]),
+                              padding=1, with_header_hide=None),
                   "rst":
                   TableFormat(lineabove=Line("", "=", "  ", ""),
                               linebelowheader=Line("", "=", "  ", ""),
@@ -110,24 +161,27 @@ _table_formats = {"simple":
                               linebelow=Line("", "=", "  ", ""),
                               headerrow=DataRow("", "  ", ""),
                               datarow=DataRow("", "  ", ""),
-                              padding=0,
-                              usecolons=False,
-                              usehtmlattrs=False,
-                              with_header_hide=[],
-                              without_header_hide=["linebelowheader"]),
+                              padding=0, with_header_hide=None),
                   "mediawiki":
                   TableFormat(lineabove=Line("{| class=\"wikitable\" style=\"text-align: left;\"",
                                              "", "", "\n|+ <!-- caption -->\n|-"),
                               linebelowheader=Line("|-", "", "", ""),
                               linebetweenrows=Line("|-", "", "", ""),
                               linebelow=Line("|}", "", "", ""),
-                              headerrow=DataRow("!", "!!", ""),
-                              datarow=DataRow("|", "||", ""),
-                              padding=1,
-                              usecolons=False,
-                              usehtmlattrs=True,
-                              with_header_hide=[],
-                              without_header_hide=["linebelowheader"])}
+                              headerrow=partial(_mediawiki_row_with_attrs, "!"),
+                              datarow=partial(_mediawiki_row_with_attrs, "|"),
+                              padding=0, with_header_hide=None),
+                  "latex":
+                  TableFormat(lineabove=_latex_line_begin_tabular,
+                              linebelowheader=Line("\\hline", "", "", ""),
+                              linebetweenrows=None,
+                              linebelow=Line("\\hline\n\\end{tabular}", "", "", ""),
+                              headerrow=DataRow("", "&", "\\\\"),
+                              datarow=DataRow("", "&", "\\\\"),
+                              padding=1, with_header_hide=None)}
+
+
+tabulate_formats = list(sorted(_table_formats.keys()))
 
 
 _invisible_codes = re.compile("\x1b\[\d*m")  # ANSI color codes
@@ -136,13 +190,14 @@ _invisible_codes = re.compile("\x1b\[\d*m")  # ANSI color codes
 def simple_separated_format(separator):
     """Construct a simple TableFormat with columns separated by a separator.
 
-    >>> tsv = simple_separated_format("\t") ; \
+    >>> tsv = simple_separated_format("\\t") ; \
         tabulate([["foo", 1], ["spam", 23]], tablefmt=tsv) == u'foo \\t 1\\nspam\\t23'
     True
 
     """
     return TableFormat(None, None, None, None,
-                       headerrow=None, datarow=DataRow('', '\t', ''), **_format_defaults)
+                       headerrow=None, datarow=DataRow('', separator, ''),
+                       padding=0, with_header_hide=None)
 
 
 def _isconvertible(conv, string):
@@ -199,6 +254,8 @@ def _type(string, has_invisible=True):
 
     if string is None:
         return _none_type
+    elif hasattr(string, "isoformat"):  # datetime.datetime, date, and time
+        return _text_type
     elif _isint(string):
         return int
     elif _isnumber(string):
@@ -345,6 +402,9 @@ def _column_type(strings, has_invisible=True):
     True
     >>> _column_type([1, 2, None]) is _int_type
     True
+    >>> import datetime as dt
+    >>> _column_type([dt.datetime(1991,2,19), dt.time(17,35)]) is _text_type
+    True
 
     """
     types = [_type(s, has_invisible) for s in strings ]
@@ -390,8 +450,12 @@ def _normalize_tabular_data(tabular_data, headers):
 
     * list-of-lists or another iterable of iterables
 
+    * list of named tuples (usually used with headers="keys")
+
     * 2D NumPy arrays
 
+    * NumPy record arrays (usually used with headers="keys")
+
     * dict of iterables (usually used with headers="keys")
 
     * pandas.DataFrame (usually used with headers="keys")
@@ -422,7 +486,17 @@ def _normalize_tabular_data(tabular_data, headers):
     else:  # it's a usual an iterable of iterables, or a NumPy array
         rows = list(tabular_data)
 
-        if headers == "keys" and len(rows) > 0:  # keys are column indices
+        if (headers == "keys" and
+            hasattr(tabular_data, "dtype") and
+            getattr(tabular_data.dtype, "names")):
+            # numpy record array
+            headers = tabular_data.dtype.names
+        elif (headers == "keys"
+              and len(rows) > 0
+              and isinstance(rows[0], tuple)
+              and hasattr(rows[0], "_fields")): # namedtuple
+            headers = list(map(_text_type, rows[0]._fields))
+        elif headers == "keys" and len(rows) > 0:  # keys are column indices
             headers = list(map(_text_type, range(len(rows[0]))))
 
     # take headers from the first row if necessary
@@ -456,8 +530,9 @@ def tabulate(tabular_data, headers=[], tablefmt="simple",
     ---  ---------
 
     The first required argument (`tabular_data`) can be a
-    list-of-lists (or another iterable or iterables), a dictionary of
-    iterables, a two-dimensional NumPy array, or a Pandas' dataframe.
+    list-of-lists (or another iterable of iterables), a list of named
+    tuples, a dictionary of iterables, a two-dimensional NumPy array,
+    NumPy record array, or a Pandas' dataframe.
 
 
     Table headers
@@ -512,7 +587,9 @@ def tabulate(tabular_data, headers=[], tablefmt="simple",
     -----  --  ----
 
     Various plain-text table formats (`tablefmt`) are supported:
-    'plain', 'simple', 'grid', 'pipe', 'orgtbl', 'rst', and 'mediawiki'.
+    'plain', 'simple', 'grid', 'pipe', 'orgtbl', 'rst', 'mediawiki',
+    and 'latex'. Variable `tabulate_formats` contains the list of
+    currently supported formats.
 
     "plain" format doesn't use any pseudographics to draw tables,
     it separates columns with a double space:
@@ -627,15 +704,15 @@ def tabulate(tabular_data, headers=[], tablefmt="simple",
     | eggs      || align="right"|  451
     |}
 
-    >>> print(tabulate([["eggs", 42], ["spam", 23]], tablefmt="mediawiki", stralign="left"))
-    {| class="wikitable" style="text-align: left;"
-    |+ <!-- caption -->
-    |-
-    | eggs || align="right"| 42
-    |-
-    | spam || align="right"| 23
-    |}
+    "latex" produces a tabular environment of LaTeX document markup:
 
+    >>> print(tabulate([["spam", 41.9999], ["eggs", "451.0"]], tablefmt="latex"))
+    \\begin{tabular}{lr}
+    \\hline
+     spam &  41.9999 \\\\
+     eggs & 451      \\\\
+    \\hline
+    \\end{tabular}
 
     """
 
@@ -679,82 +756,74 @@ def tabulate(tabular_data, headers=[], tablefmt="simple",
     return _format_table(tablefmt, headers, rows, minwidths, aligns)
 
 
-def _build_row(cells, padding, begin, sep, end):
-    "Return a string which represents a row of data cells."
-    pad = u" "*padding
-    padded_cells = [pad + cell + pad for cell in cells]
+def _build_simple_row(padded_cells, rowfmt):
+    "Format row according to DataRow format without padding."
+    begin, sep, end = rowfmt
     return (begin + sep.join(padded_cells) + end).rstrip()
 
 
-def _build_line(colwidths, padding, begin, fill, sep,  end):
-    "Return a string which represents a horizontal line."
-    cells = [fill*(w + 2*padding) for w in colwidths]
-    return _build_row(cells, 0, begin, sep, end)
-
-
-def _mediawiki_cell_attrs(row, colaligns):
-    "Prefix every cell in a row with an HTML alignment attribute."
-    alignment = { "left":    '',
-                  "right":   'align="right"| ',
-                  "center":  'align="center"| ',
-                  "decimal": 'align="right"| ' }
-    row2 = [alignment[a] + c for c, a in zip(row, colaligns)]
-    return row2
-
-
-def _line_segment_with_colons(linefmt, align, colwidth):
-    """Return a segment of a horizontal line with optional colons which
-    indicate column's alignment (as in `pipe` output format)."""
-    fill = linefmt.hline
-    w = colwidth
-    if align in ["right", "decimal"]:
-        return (fill[0] * (w - 1)) + ":"
-    elif align == "center":
-        return ":" + (fill[0] * (w - 2)) + ":"
-    elif align == "left":
-        return ":" + (fill[0] * (w - 1))
+def _build_row(padded_cells, colwidths, colaligns, rowfmt):
+    "Return a string which represents a row of data cells."
+    if not rowfmt:
+        return None
+    if hasattr(rowfmt, "__call__"):
+        return rowfmt(padded_cells, colwidths, colaligns)
     else:
-        return fill[0] * w
+        return _build_simple_row(padded_cells, rowfmt)
+
+
+def _build_line(colwidths, colaligns, linefmt):
+    "Return a string which represents a horizontal line."
+    if not linefmt:
+        return None
+    if hasattr(linefmt, "__call__"):
+        return linefmt(colwidths, colaligns)
+    else:
+        begin, fill, sep,  end = linefmt
+        cells = [fill*w for w in colwidths]
+        return _build_simple_row(cells, (begin, sep, end))
+
+
+def _pad_row(cells, padding):
+    if cells:
+        pad = u" "*padding
+        padded_cells = [pad + cell + pad for cell in cells]
+        return padded_cells
+    else:
+        return cells
 
 
 def _format_table(fmt, headers, rows, colwidths, colaligns):
     """Produce a plain-text representation of the table."""
     lines = []
-    hidden = fmt.with_header_hide if headers else fmt.without_header_hide
+    hidden = fmt.with_header_hide if (headers and fmt.with_header_hide) else []
     pad = fmt.padding
-    headerrow = fmt.headerrow if fmt.headerrow else fmt.datarow
+    headerrow = fmt.headerrow
 
-    if fmt.usehtmlattrs:
-        headers = _mediawiki_cell_attrs(headers, colaligns)
-        rows = [_mediawiki_cell_attrs(row, colaligns) for row in rows]
+    padded_widths = [(w + 2*pad) for w in colwidths]
+    padded_headers = _pad_row(headers, pad)
+    padded_rows = [_pad_row(row, pad) for row in rows]
 
     if fmt.lineabove and "lineabove" not in hidden:
-        lines.append(_build_line(colwidths, pad, *fmt.lineabove))
+        lines.append(_build_line(padded_widths, colaligns, fmt.lineabove))
 
-    if headers:
-        lines.append(_build_row(headers, pad, *headerrow))
+    if padded_headers:
+        lines.append(_build_row(padded_headers, padded_widths, colaligns, headerrow))
+        if fmt.linebelowheader and "linebelowheader" not in hidden:
+            lines.append(_build_line(padded_widths, colaligns, fmt.linebelowheader))
 
-    if fmt.linebelowheader and "linebelowheader" not in hidden:
-        begin, fill, sep, end = fmt.linebelowheader
-        if fmt.usecolons:
-            segs = [_line_segment_with_colons(fmt.linebelowheader, a, w + 2*pad)
-                    for w,a in zip(colwidths, colaligns)]
-            lines.append(_build_row(segs, 0, begin, sep, end))
-        else:
-            lines.append(_build_line(colwidths, pad, *fmt.linebelowheader))
-
-    if rows and fmt.linebetweenrows and "linebetweenrows" not in hidden:
+    if padded_rows and fmt.linebetweenrows and "linebetweenrows" not in hidden:
         # initial rows with a line below
-        for row in rows[:-1]:
-            lines.append(_build_row(row, pad, *fmt.datarow))
-            lines.append(_build_line(colwidths, pad, *fmt.linebetweenrows))
+        for row in padded_rows[:-1]:
+            lines.append(_build_row(row, padded_widths, colaligns, fmt.datarow))
+            lines.append(_build_line(padded_widths, colaligns, fmt.linebetweenrows))
         # the last row without a line below
-        lines.append(_build_row(rows[-1], pad, *fmt.datarow))
+        lines.append(_build_row(padded_rows[-1], padded_widths, colaligns, fmt.datarow))
     else:
-        for row in rows:
-            lines.append(_build_row(row, pad, *fmt.datarow))
+        for row in padded_rows:
+            lines.append(_build_row(row, padded_widths, colaligns, fmt.datarow))
 
     if fmt.linebelow and "linebelow" not in hidden:
-        lines.append(_build_line(colwidths, pad, *fmt.linebelow))
+        lines.append(_build_line(padded_widths, colaligns, fmt.linebelow))
 
     return "\n".join(lines)
